import json
import pickle
from pathlib import Path
from typing import Dict, Sized

import networkx as nx

from vizitig.errors import EmptyExportError, FileExists
from vizitig.info import get_graph
from vizitig.types import Kmer, Color
from vizitig.utils import vizitig_logger

export_format = dict(
    bcalm="fa",
    json="json",
    networkx_pickle="pickle",
    gml="gml",
    nodelist="list",
)


def export_bcalm_by_iterator(graph_name: str, outputfile: str, batch_size=100):
    G = get_graph(graph_name)
    it = G.nbunch_iter()

    somme = sum(1 for _ in it)
    print("Number of nodes: ", somme)

    it = G.nbunch_iter()
    with open(outputfile, "w") as file:
        buffer = list()
        for i, node in enumerate(it):
            buffer.append(dump_one_node_bcalm_for_iterator(G, node) + "\n")

            if (i + 1) % batch_size == 0:
                print(i)
                file.writelines(buffer)
                del buffer
                buffer = list()
                file.flush()

            del node

        if buffer:
            file.writelines(buffer)
            file.flush()


def dump_one_node_bcalm_for_iterator(G, node) -> str:
    s = f">{node}"
    sequence = G.nodes[node]["sequence"]
    neighbors: Dict[int, str] = dict()
    for target, esign in G[node].items():
        if esign:
            neighbors[target] = esign["sign"].value
        else:
            neighbors[target] = (
                "++"  # compute_sign(sequence, G.nodes[target]["sequence"])
            )
    edges = (f"L:{sign[0]}:{target}:{sign[1]}" for target, sign in neighbors.items())
    s += f" {' '.join(edges)}"

    colors: list[str] = list()
    for item in G.nodes[node]:
        if isinstance(item, Color):
            colors.append(str(item.offset))
    colors.reverse()
    color_header = ":".join(colors)
    return f"{s} [{color_header}]\n{sequence}"


def dump_one_node_bcalm(G, node) -> str:
    s = f">{node}"
    sequence = G.nodes[node]["sequence"]
    neighbors: Dict[int, str] = dict()
    for target, esign in G[node].items():
        if esign:
            neighbors[target] = esign["sign"].value
        else:
            neighbors[target] = (
                "  "  # compute_sign(sequence, G.nodes[target]["sequence"])
            )
    edges = (f"L:{sign[0]}:{target}:{sign[1]}" for target, sign in neighbors.items())
    s += f" {' '.join(edges)}"
    return f"{s}\n{sequence}"


def dump_bcalm(graph: nx.Graph, output_file: Path):
    with open(output_file, "a") as f:
        f.write("\n".join(map(lambda node: dump_one_node_bcalm(graph, node), graph)))


def export_graph(
    graph_name: str, nodes: Sized, format: str, output: Path, bcalm_iterator=False
):
    # Get a clean path object
    output_file = Path(output)

    # Raise en error if the file already exists
    if output_file.exists():
        raise FileExists(output_file)

    # Get the graph

    Graph = get_graph(graph_name)

    # Create a subgraph with the nodes to store
    # The subgraph will be loaded in RAM by networkx
    SubGraph = Graph.subgraph(nodes, temporary_table=True).copy_to_networkx()
    if len(SubGraph) < len(nodes):
        vizitig_logger.warning(
            f"Some nodes are not experted (nodes exported: {len(SubGraph)}/{len(nodes)})",
        )
    if len(SubGraph) == 0:
        raise EmptyExportError

    for node, data in SubGraph.nodes(data=True):
        for key in list(data):
            if isinstance(key, Kmer):
                del data[key]
            if hasattr(key, "short_repr"):
                val = data.pop(key)
                short_repr = key.short_repr()
                if format == "gml":
                    short_repr = short_repr.replace("(", "").replace(")", "")
                data[short_repr] = val

    format = format.lower()

    if format == "json":
        with open(output_file, "w") as f:
            json.dump(nx.node_link_data(SubGraph), f)
    elif format == "networkx_pickle":
        with open(output_file, "wb") as f:
            pickle.dump(SubGraph, f)
    elif format == "bcalm":
        dump_bcalm(SubGraph, output_file)
    elif format == "gml":
        SubGraph.graph = {}

        nx.write_gml(SubGraph, output_file, lambda e: "" if e is None else e)
    elif format == "nodelist":
        with open(output_file, "w") as f:
            f.write("\n".join(map(str, SubGraph)))
    else:
        raise NotImplementedError(format)



from os import getenv, cpu_count
from pathlib import Path
import psutil
import platform

# VIZITIG_TMP_DIR:
#   directory to store temp files, Default None
VIZITIG_TMP_DIR = getenv("VIZITIG_TMP_DIR")

# VIZITIG_PROC_NUMBER:
#   default number of shard when building index,
#   default  min(cpu_c - 1, 10) where cpu_c is the CPU_COUNT

if getenv("VIZITIG_PROC_NUMBER"):
    VIZITIG_PROC_NUMBER = int(str(getenv("VIZITIG_PROC_NUMBER")))
else:
    cpu_c = cpu_count()
    if cpu_c is not None:
        VIZITIG_PROC_NUMBER = min(cpu_c - 1, 10)
    else:
        VIZITIG_PROC_NUMBER = 4

if getenv("VIZITIG_WORK_MEM"):
    VIZITIG_WORK_MEM = int(str(getenv("VIZITIG_WORK_MEM")))
else:
    VIZITIG_WORK_MEM = int(0.6 * psutil.virtual_memory().total)


# VIZITIG_DIR:
#   directory where to store vizitig data, default ~/.vizitig

VIZITIG_DIR = Path(getenv("VIZITIG_DIR", "~/.vizitig")).expanduser()


# VIZITIG_PYTHON_ONLY:
#   use only Python types and do not try to use vizibridge, default False

VIZITIG_PYTHON_ONLY = getenv("VIZITIG_PYTHON_ONLY", False)


# VIZITIG_DEFAULT_INDEX:
#   set the default index to use in vizitig. Default False
#   if not set, vizitig will choose one.

VIZITIG_DEFAULT_INDEX = getenv("VIZITIG_DEFAULT_INDEX")

# VIZITIG_NO_TMP_INDEX:
#   if set, will build temporary index

VIZITIG_NO_TMP_INDEX = False
if getenv("VIZITIG_NO_TMP_INDEX"):
    VIZITIG_NO_TMP_INDEX = True

# VIZITIG_SHORT_TEST:
#   if set, will do less test

VIZITIG_SHORT_TEST = False
if getenv("VIZITIG_SHORT_TEST"):
    VIZITIG_SHORT_TEST = True

# VIZITIG_NO_PARALLEL_INDEX:
#   if set, do not build the index using python multiprocessing
#   default False for linux and True for the rest

VIZITIG_NO_PARALLEL_INDEX = True
if platform.system() == "Linux":
    VIZITIG_NO_PARALLEL_INDEX = False
if getenv("VIZITIG_NO_PARALLEL_INDEX"):
    VIZITIG_NO_PARALLEL_INDEX = True



import sys

from vizitig import annotate, generate_graph, index, info, run, color
from vizitig.cli import parser

__all__ = ["info", "generate_graph", "run", "color", "annotate", "index"]


def main():
    args = parser.parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()



import argparse
from pathlib import Path
from typing import Iterator, Mapping, MutableMapping, Tuple
from tempfile import NamedTemporaryFile
import shutil

from vizitig.cli import subparsers
from vizitig.errors import GraphExists
from vizitig.info import GraphLogger, create_graph, graph_path_name
from vizitig.metadata import GraphMetadata
from vizitig.parsing import NodeHeader, parse, stat_bcalm
from vizitig.utils import (
    IteratableFromGenerator,
    SubLog,
    progress,
    vizitig_logger as logger,
)
from vizitig.env_var import VIZITIG_TMP_DIR


def nodes_iter_from_desc(
    it: Iterator[NodeHeader],
    k: int,
    graph_metadata: GraphMetadata,
    kmer_ingestion: bool,
) -> Iterator[Tuple[int, Mapping]]:
    for node in it:
        d: MutableMapping = dict(sequence=str(node.sequence))
        # HERE : add metadata
        if kmer_ingestion:
            for i, kmer in enumerate(node.sequence.enum_canonical_kmer(k)):
                d[kmer] = i
        if node.occurence:
            d["occurence"] = node.occurence
        # d must follow carefully the metadata data plan
        yield (node.node_id, d)


def generate_graph(
    path: Path,
    output_path: Path,
    name: str,
    k: int | None = None,
    kmer_ingestion: bool = False,
    buffsize: int = 10**6,  # Why this value ?
    parse_only: bool = False,
    sql_logger: bool = False,
    edge_annotation: bool = False,
    directed_edges: bool = True,
    raise_no_edges: bool = False,
) -> None:
    """Generate a graph file for vizitig to use"""
    if output_path.exists():
        raise GraphExists(output_path.stem)

    with GraphLogger(name):
        if parse_only:
            logger.info("Parse only execution")

        # Compute the number of nodes
        # Uses size_bcalm from vizitig/parsing
        logger.info("Computing some stat")
        size, edge_size, estimate_k, kmer_size = stat_bcalm(path, buffsize=buffsize)
        logger.info(
            f"found {size} nodes, {edge_size} edges and estimated k {estimate_k} ({k} provided)",
        )
        if edge_size == 0:
            msg = "No edge found -- is the input format BCALM?"
            if raise_no_edges:
                raise ValueError(msg)
            logger.info(msg)

        if k is None:
            k = estimate_k

        if not parse_only:
            G = create_graph(
                name,
                k,
                output_path,
                size=size,
                kmer_size=kmer_size,
                edge_size=edge_size,
                sql_logger=sql_logger,
                directed_edges=directed_edges,
            )
            GM = G.metadata
            G.drop_index()
            G.helper.pragma_fk = False
        logger.info(f"Loading nodes, {kmer_ingestion=}")
        # Builds the list of node (nodeid, dict (canonical kmer, sequence))
        nodes_it = IteratableFromGenerator(
            lambda: progress(
                nodes_iter_from_desc(
                    parse(path, buffsize=buffsize),
                    k,
                    graph_metadata=GM,
                    kmer_ingestion=kmer_ingestion,
                ),
                total=size,
            ),
        )

        if not parse_only:
            with SubLog("node_ingestion"):
                G.add_nodes_from(nodes_it)

        logger.info("Loading edges")
        desc = parse(path, buffsize=buffsize)
        if edge_annotation:

            def edges_it_annotate():
                return progress(
                    (
                        (node.node_id, f, {"sign": s.value})
                        for node in desc
                        for f, s in node.successors
                    ),
                    total=edge_size,
                )

            if not parse_only:
                with SubLog("edge_ingestion"):
                    G.add_edges_from(IteratableFromGenerator(edges_it_annotate))
        else:

            def edges_it():
                return progress(
                    ((node.node_id, f) for node in desc for f, _ in node.successors),
                    total=edge_size,
                )

            with SubLog("edge_ingestion"):
                G.add_edges_without_data_from(IteratableFromGenerator(edges_it))

        if not parse_only:
            with SubLog("indexing"):
                G.reindex()


def build_cli(args):
    """Builds a solid generate_graph call
    Tests if the args are correct and fetches the global variables such as path
    """
    # Name of the input
    input = Path(args.filename)

    # Raises an error if the file does not exist
    if not input.exists():
        raise ValueError(f"{input} is not a valid path to a file")

    name = args.name
    if not name:
        name = input.name

    output = graph_path_name(name)

    # Deletes the current output if already existants and erase = store_true
    # Raises an error otherwise
    if output.exists():
        if args.erase:
            output.unlink()
        else:
            raise GraphExists(output.stem)
    # Finally generate the graph
    with SubLog("build"):
        file = NamedTemporaryFile(prefix=VIZITIG_TMP_DIR)
        file.close()
        tmp_output = Path(file.name)
        generate_graph(
            input,
            tmp_output,
            name,
            args.k,
            buffsize=args.buffer_size,
            parse_only=args.parse_only,
            kmer_ingestion=args.kmer_ingestion,
            sql_logger=args.sql_logger,
            directed_edges=args.directed_edges,
            edge_annotation=args.edge_annotation,
        )
        logger.info(f"moving {tmp_output} to {input}")
        shutil.move(tmp_output, output)


build = subparsers.add_parser(
    "build",
    help="build a new graph from BCALM file",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
build.add_argument(
    "filename",
    help="A path to BCALM fize (gz compressed possibly)",
    metavar="filename",
    type=str,
)
build.add_argument(
    "-k",
    "--k",
    help="Size of the kmers used to build the graph. By default, the k is estimated from the input file",
    metavar="k",
    type=int,
)

build.add_argument(
    "-e",
    "--erase",
    help="Delete previous graph-file if it already exists",
    action="store_true",
)

build.add_argument(
    "-n",
    "--name",
    help="Set the name of the graph, default is input filename",
)

build.add_argument(
    "--sql-logger",
    help="display the SQL logger",
    action="store_true",
)

build.add_argument(
    "-p",
    "--parse-only",
    help="Simply parse the file and iterates through it without storing the result. For benchmarking purpose",
    action="store_true",
)

build.add_argument(
    "-b",
    "--buffer-size",
    help="A parameter used to bound the memory usage. In case of memory exhaustion, lower the value and increase it to get a faster processing",
    type=int,
    default=10**4,
)

build.add_argument(
    "-i",
    "--kmer-ingestion",
    help="Flag to ingestion kmers. Not ingesting kmers reduces the build time and the size of the vizitig graph, but make the fetching of kmer or require the build of an extra index",
    action="store_true",
    default=False,
)

build.add_argument(
    "--edge-annotation",
    help="Add sign annotation to edges (--, -+, +-, ++)",
    action="store_true",
    default=False,
)

build.add_argument(
    "-d",
    "--directed-edges",
    help="Store edge orientation",
    action="store_true",
    default=False,
)


build.set_defaults(func=build_cli)



from fastapi import HTTPException

from vizitig import version as viz_version
from vizitig.errors import NoGraphException as InfoNoGraphException


class NoGraphException(InfoNoGraphException, HTTPException):
    def __init__(self, **kwargs):
        detail = kwargs.pop("detail", "")
        super().__init__(detail=f'No Graph "{detail}"', **kwargs)


class NotFoundException(HTTPException):
    _object = "Something"

    def __init__(self, **kwargs):
        detail = kwargs.pop("detail", "")
        super().__init__(detail=f'{self._object} not found: "{detail}"', **kwargs)


class NodeNotFoundException(NotFoundException):
    _object = "Node"


class MetaDataNotFoundException(NotFoundException):
    _object = "Metadata key"


class NoPathFoundException(NotFoundException):
    _object = "No path found"


class UnknownFormat(NotFoundException):
    _object = "Format not found"


class ExportError(NotFoundException):
    _object = "Error during export"


class ReservedKeysException(HTTPException):
    def __init__(self, **kwargs):
        detail = kwargs.pop("detail", "")
        super().__init__(detail=f'Reserved keys "{detail}"', **kwargs)


class InvalidKmerSize(HTTPException):
    def __init__(self, **kwargs):
        seq = kwargs.pop("seq", "")
        k = kwargs.pop("k", "")
        detail = f"{seq} ({len(seq)}!={k})" if seq and k else ""
        super().__init__(detail=f'Invalid Kmersize: "{detail}"', **kwargs)


class QueryError(HTTPException):
    _object = "Error while parsing"


class GraphNameAlreayTaken(HTTPException):
    def __init__(self, **kwargs):
        name = kwargs.pop("name")
        super().__init__(detail=f"Graph name {name} already taken", status_code=400)


class NoEmptyGraphNameAllowed(HTTPException):
    _object = "No empty graph name allowed"


class UnsupportedExtension(HTTPException):
    def __init__(self, **kwargs):
        ext = kwargs.pop("extension")
        super().__init__(detail=f"Unsupported extension {ext}", status_code=400)


class IncompatibleViziGraph(HTTPException):
    def __init__(self, **kwargs):
        version = kwargs.pop("version")
        super().__init__(
            detail=f"Incompatible vizitig version ({version} vs {viz_version})",
        )



from typing import Literal, TypeAlias

from lark import Lark, Transformer
from lark import exceptions as larkexceptions
from pydantic.dataclasses import dataclass
from vizitig.errors import ParseError


class Base:
    @property
    def t(self):
        keys = self.__dataclass_fields__.values()
        key = next(iter(keys))
        return getattr(self, key.name)


@dataclass(frozen=True)
class All(Base):
    all: None = None


@dataclass(frozen=True)
class Loop(Base):
    loop: None = None


@dataclass(frozen=True)
class Partial(Base):
    partial: None = None


@dataclass(frozen=True)
class Selection(Base):
    selection: None = None


@dataclass(frozen=True)
class NodeId(Base):
    id: tuple[int, ...]


@dataclass(frozen=True)
class Attr:
    key: str
    value: str | int
    op: Literal["=", "<", ">"]


@dataclass(frozen=True)
class Meta(Base):
    type: str
    name: str | None = None
    attrs: tuple[Attr, ...] = ()


@dataclass(frozen=True)
class Op(Base):
    operation: Literal["=", "<", ">", "<=", ">="]


@dataclass(frozen=True)
class Abundance(Base):
    operation: Op
    value: int


@dataclass(frozen=True)
class Color(Base):
    color: str
    abundance: Abundance | None = None


@dataclass(frozen=True)
class Kmer(Base):
    kmer: str


@dataclass(frozen=True)
class Degree(Base):
    degree: int


@dataclass(frozen=True)
class Threshold(Base):
    val: int


@dataclass(frozen=True)
class SmallK(Base):
    val: int


@dataclass(frozen=True)
class Seq(Base):
    seq: str
    threshold: Threshold | None = None
    smallk: SmallK | None = None


@dataclass(frozen=True)
class PseudoMapping(Base):
    psdomap: str


@dataclass(frozen=True)
class And(Base):
    land: list["Term"]


@dataclass(frozen=True)
class Or(Base):
    lor: list["Term"]


@dataclass(frozen=True)
class Not(Base):
    lnot: "Term"


@dataclass(frozen=True)
class Parenthesis(Base):
    par: "Term"


Term: TypeAlias = (
    NodeId
    | Meta
    | Color
    | And
    | Or
    | Not
    | Kmer
    | str
    | Seq
    | Op
    | Abundance
    | All
    | Loop
    | Partial
    | Selection
    | Degree
    | Parenthesis
)


Grammar = r"""
?formula: lor_infix  
        | lor_prefix
        | land 
        | lnot 
        | par  
        | literal

_LEFTPAR  : /\s*\(\s*/ 
_RIGHTPAR : /\s*\)\s*/
par      : _LEFTPAR formula _RIGHTPAR 
_SEP      : /\s*,\s*/
lor_infix : formula "or"i (_SPACES formula | par) 
lor_prefix: "or"i _LEFTPAR (formula ( _SEP formula )*)? _RIGHTPAR
land      : formula "and"i (_SPACES formula | par)
lnot      : "not"i (_SPACES formula | par)
_SPACES   : /\s+/
        
?literal: nodeid 
        | kmer 
        | color
        | psdomap
        | meta
        | seq
        | all
        | loop
        | partial
        | selection 
        | degree 
    
selection.1: "selection"i ((_LEFTPAR _RIGHTPAR) |)
partial.1  : "Partial"i ((_LEFTPAR _RIGHTPAR) |)
all.1      : "All"i ((_LEFTPAR _RIGHTPAR) |)
loop.1      : "loop"i ((_LEFTPAR _RIGHTPAR) |)
nodeid.1   : "NodeId"i _LEFTPAR  integer ("," integer)* _RIGHTPAR
color.1    : "Color"i _LEFTPAR  ident ("," abundance |)  _RIGHTPAR
degree.1   : "Degree"i _LEFTPAR integer _RIGHTPAR
meta.0     : ident _LEFTPAR (arg ( _SEP  arg )* | ) _RIGHTPAR
?arg       : ident | attr

abundance  : "A" op integer
kmer.2     : "Kmer"i _LEFTPAR acgt  _RIGHTPAR
attr       : ident op (ident | integer)
seq.1      : "Seq"i _LEFTPAR acgt ( "," threshold)? ("," small_k)?  _RIGHTPAR
psdomap.1  : ("PseudoMapping"i | "PM"i ) _LEFTPAR acgt _RIGHTPAR
?op        : (EQUAL | LT | GT | LTE | GTE)

EQUAL      : "="
LT         : "<"
GT         : ">"
LTE        : "<="
GTE        : ">="
threshold : ("T"i | "Threshold"i) "=" integer
small_k    : ("k"i | "small-k"i) "="  integer

acgt       : /[ACGT]+/i 
integer    : INT
ident      : /[a-zA-Z_][\w\.\-\_]*/i
    
%import common.CNAME
%import common.INT
%import common.WS 
%ignore WS
"""


class QueryEval(Transformer):
    def kmer(self, e):
        return Kmer(e[0])

    def degree(self, e):
        return Degree(e[0])

    def seq(self, e):
        kwargs = dict()
        for f in e[1:]:
            if isinstance(f, SmallK):
                kwargs["smallk"] = f
            if isinstance(f, Threshold):
                kwargs["threshold"] = f
        return Seq(e[0], **kwargs)

    def all(self, *args):
        return All()

    def loop(self, *args):
        return Loop()

    def partial(self, *args):
        return Partial()

    def selection(self, *args):
        return Selection()

    def acgt(self, e):
        # return next(DNA.from_str(e[0]))
        return e[0]

    def psdomap(self, e):
        return PseudoMapping(e[0])

    def integer(self, e):
        return int(e[0])

    def ident(self, e):
        return str(e[0])

    def threshold(self, e):
        return Threshold(e[0])

    def small_k(self, e):
        return SmallK(e[0])

    def abundance(self, e):
        return Abundance(Op(e[0]), e[1])

    def color(self, e):
        return Color(*e)

    def attr(self, e):
        return Attr(key=e[0], op=e[1], value=e[2])

    def op(self, e):
        return str(e[0])

    def meta(self, args):
        type = args[0]
        name = None
        attrs = []
        for arg in args[1:]:
            if isinstance(arg, Attr):
                attrs.append(arg)
                continue
            assert name is None
            name = arg
        return Meta(type=type, name=name, attrs=attrs)

    def nodeid(self, e):
        return NodeId(e)

    def land(self, e):
        return And(list(filter(bool, e)))

    def lor(self, e):
        return Or(list(filter(bool, e)))

    lor_prefix = lor
    lor_infix = lor

    def lnot(self, e):
        k = next(filter(bool, e))
        return Not(k)

    def par(self, e):
        k = next(filter(bool, e))
        return Parenthesis(k)


parser = Lark(Grammar, start="formula")


def parse(query: str) -> Term:
    try:
        tree = parser.parse(query)
    except larkexceptions.LarkError as E:
        raise ParseError(E)
    q = QueryEval().transform(tree)
    return q



from pytest import mark
from vizibridge import DNA, Kmer

from vizitig.info import get_graph
from vizitig.types import DNAPython, KmerPython

mixed_types_kmers = [KmerPython, Kmer]
mixed_types_dna = [DNAPython, DNA]

sequences = [
    "TGCTATCGATTCGATATCAGATTCGATCGG",
    "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT",
]


class TestVizibridge:
    @mark.parametrize("seq", sequences)
    def test_vizibridge_kmers(self, seq: str, viz_dir, test_graph_name):
        G = get_graph(test_graph_name)
        k = G.metadata.k
        l1 = list(next(DNA.from_str(seq)).enum_canonical_kmer(k))
        l2 = list(next(DNAPython.from_str(seq)).enum_canonical_kmer(k))
        assert len(l1) == len(l2)
        for i in range(len(l1)):
            assert l1[i].data == l2[i].data

    @mark.parametrize("seq", sequences)
    def test_vizibridge_kmers_enumeration(self, seq: str, viz_dir, test_graph_name):
        G = get_graph(test_graph_name)
        k = G.metadata.k
        l1 = list(next(DNA.from_str(seq)).enum_kmer(k))
        l2 = list(next(DNAPython.from_str(seq)).enum_kmer(k))
        assert len(l1) == len(l2)
        for i in range(len(l1)):
            assert l1[i].data == l2[i].data



import logging
from contextlib import contextmanager
from datetime import datetime
from itertools import groupby
from math import sqrt
from pathlib import Path
from typing import Tuple
from uuid import uuid4

src_root = Path(__file__).parent

last_call = None


def chunk(g, n):
    """Return an iterable over chunk of g of size n; last chunck is of size possibly less than n."""
    for _, x in groupby(enumerate(g), lambda e: e[0] // n):
        yield map(lambda e: e[1], x)


def sizeof_fmt(num: float | int, suffix: str = "B") -> str:
    """Taken from https://stackoverflow.com/questions/1094841/get-human-readable-version-of-file-size"""
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def number_fmt(num: int) -> str:
    if num > 10**9:
        return f"{num / 10**9:0.1f}G"
    if num > 10**6:
        return f"{num / 10**6:0.1f}M"
    if num > 10**3:
        return f"{num / 10**6:0.1f}K"
    return str(num)


def new_gid():
    return str(uuid4()).replace("-", "")


def compute_sign(seq1: str, seq2: str):
    raise NotImplementedError


def cantor_pairing(k1: int, k2: int) -> int:
    """Cantor pairing function to map two non-negative integers to a single non-negative integer."""
    return (k1 + k2) * (k1 + k2 + 1) // 2 + k2


def inverse_cantor_pairing(z: int) -> Tuple[int, int]:
    """Inverse of the Cantor pairing function to map a single non-negative integer to a pair of non-negative integers."""
    w = (sqrt(8 * z + 1) - 1) // 2
    t = w * (w + 1) // 2
    k2 = z - t
    k1 = w - k2
    return (int(k1), int(k2))


class IteratableFromGenerator:
    def __init__(self, generator):
        self.generator = generator

    def __iter__(self):
        return self.generator()


# Logging utility

log_time: dict[str, datetime] = dict()
old_factory = logging.getLogRecordFactory()


def record_factory(*args, **kwargs):
    record = old_factory(*args, **kwargs)
    record.trace = ""
    record.depth = ""
    base_time = log_time.setdefault(record.trace, datetime.now())
    record.uptime = record.total_uptime = str(datetime.now() - base_time).split(".")[0]
    return record


base_fmt = "{levelname}::{asctime}::{depth}\t::{uptime}/{total_uptime}\t:: {message}\t:: {trace}"


def reset_logfmt():
    logging.basicConfig(format=base_fmt, style="{", force=True)


logging.setLogRecordFactory(record_factory)
vizitig_logger = logging.getLogger("vizitig")
vizitig_logger.setLevel("DEBUG")
reset_logfmt()


def progress(iterator, total: int | None = None):
    last = 0.0
    step = 1.0
    total = total or getattr(iterator, "__len__", lambda: None)()
    with SubLog("progress"):
        if total:
            if total < 100:
                step = 20.0
            elif total < 200:
                step = 10.0
            elif total < 400:
                step = 5.0
            else:
                step = 1.0

            for i, x in enumerate(iterator):
                yield x
                if last + step < (100 * i / total):
                    last = 100 * i / total
                    vizitig_logger.info(f"{last:0.1f}%")

        else:
            for i, x in enumerate(iterator):
                yield x
                if i % 10000 == 0:
                    vizitig_logger.info(f"{i // 1000}k")


@contextmanager
def SubLog(key: str, file: Path | None = None, delete_file=False):
    old_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        record.trace = f"{record.trace}{key}."
        base_time = log_time.setdefault(record.trace, datetime.now())
        if key.startswith("main"):
            log_time["vizitig"] = log_time[record.trace]
        if record.msg == "done":
            del log_time[record.trace]
        now = datetime.now()
        record.uptime = str(now - base_time).split(".")[0]
        record.depth += "-"
        return record

    logging.setLogRecordFactory(record_factory)
    if file:
        handler = logging.FileHandler(file)
        handler.setFormatter(logging.Formatter(base_fmt, style="{"))
        vizitig_logger.addHandler(handler)
    vizitig_logger.info("start")
    try:
        yield file
    finally:
        vizitig_logger.info("done ", extra=dict(end=True))
        logging.setLogRecordFactory(old_factory)
        if file:
            vizitig_logger.removeHandler(handler)



import re
import math
from dataclasses import dataclass
from enum import Enum
from typing import Iterator, List, Literal, Mapping, SupportsIndex, cast

from typing_extensions import Self

from vizitig.env_var import VIZITIG_PYTHON_ONLY
from vizitig.utils import vizitig_logger as log

Nucleotide = Literal["A", "C", "G", "T"]
non_CGTA = re.compile("[^ACGT]")


class ESign(Enum):
    pp = "++"
    mp = "-+"
    mm = "--"
    pm = "+-"


class DNAPython(str):
    """A light class to type string with only letters ACGT"""

    def __iter__(self) -> Iterator[Nucleotide]:
        for x in super().__iter__():
            yield cast(Nucleotide, x)

    def enum_kmer(self, k: int) -> Iterator["Kmer"]:
        if len(self) < k:
            return
        kmer = KmerPython.from_sequence(self[:k])
        yield kmer
        for a in self[k:]:
            kmer = kmer.add_right_nucleotid(a)
            yield kmer

    def enum_canonical_kmer(self, k: int) -> Iterator["Kmer"]:
        if len(self) < k:
            return
        kmer = KmerPython.from_sequence(self[:k])
        rc = kmer.reverse_complement()
        yield min(kmer, rc)
        for a in self[k:]:
            kmer = kmer.add_right_nucleotid(a)
            rc = rc.add_left(complement_table[char_to_int[a]])
            yield min(kmer, rc)

    @classmethod
    def from_str(cls, seq: str) -> Iterator["DNA"]:
        yield from (cls(subseq) for subseq in non_CGTA.split(seq))

    def __getitem__(self, __key: SupportsIndex | slice) -> Self:
        return type(self)(super().__getitem__(__key))


Quarter = Literal[0b00, 0b01, 0b10, 0b11]

char_to_int: Mapping[Nucleotide, Quarter] = {"A": 0b00, "C": 0b01, "G": 0b10, "T": 0b11}
int_to_char: Mapping[Quarter, Nucleotide] = {0b00: "A", 0b01: "C", 0b10: "G", 0b11: "T"}
complement_table: Mapping[Quarter, Quarter] = {
    k: cast(Quarter, (~k & 0b11)) for k in int_to_char
}


if not VIZITIG_PYTHON_ONLY:
    try:
        from vizibridge import DNA
    except ImportError:
        DNA = DNAPython
else:
    DNA = DNAPython


@dataclass
class KmerPython:
    data: int
    size: int

    @classmethod
    def from_sequence(cls, seq: DNA) -> Self:
        return cls.from_iter(map(char_to_int.__getitem__, seq))  # type: ignore
        # mypy claim that it expect a str instead of a Nucleotide ???

    def __reduce__(self):
        return (self.__class__, (self.data, self.size))

    @classmethod
    def from_iter(cls, it: Iterator[Quarter]) -> Self:
        data = 0
        size = 0
        for q in it:
            data += q
            data = data << 2
            size += 1
        return cls(data >> 2, size)

    def __iter__(self):
        data = self.data
        for a in range(self.size)[::-1]:
            yield (data & (0b11 << (a * 2))) >> (a * 2)

    def __repr__(self):
        return "".join(map(int_to_char.__getitem__, self))

    def add_left_nucleotid(self, n: Nucleotide) -> Self:
        return self.add_left(char_to_int[n])

    def add_right_nucleotid(self, n: Nucleotide) -> Self:
        return self.add_right(char_to_int[n])

    def add_left(self, n: Quarter) -> Self:
        data = (self.data >> 2) + (n << ((self.size - 1) * 2))
        return type(self)(data, self.size)

    def add_right(self, n: Quarter) -> Self:
        data = ((self.data << 2) + (n)) & ((1 << (2 * self.size)) - 1)
        return type(self)(data, self.size)

    def reverse_complement(self) -> Self:
        c = map(complement_table.__getitem__, reversed(list(self)))  # type: ignore
        return self.from_iter(c)  # type: ignore
        # mypy claim weird stuff ...

    def canonical(self) -> Self:
        return min(self, self.reverse_complement())

    def is_canonical(self) -> bool:
        return self == self.canonical()

    def __hash__(self):
        return hash((self.data, self.size))

    def __lt__(self, other) -> bool:
        assert self.size == other.size
        return self.data <= other.data

    def __gt__(self, other) -> bool:
        assert self.size == other.size
        return self.data >= other.data


if not VIZITIG_PYTHON_ONLY:
    try:
        from vizibridge import Kmer, KmerTypeMap

        # Here Kmer is a only a type (a union of classes) and can't be instantiated directly.
        AvailableKmerTypes = KmerPython | Kmer
        AvailableKmerSize = set(KmerTypeMap)
    except ImportError as E:
        log.warning(f"Error importing Rust backend {E}")

        AvailableKmerTypes = KmerPython
        Kmer = KmerPython
        AvailableKmerSize = set(range(2, 63))
else:
    AvailableKmerTypes = KmerPython
    Kmer = KmerPython
    AvailableKmerSize = set(range(2, 63))


@dataclass
class Color:
    id: str
    description: str
    type: str = "Color"
    offset: int | None = None

    def set_offset(self, offset: int):
        self.offset = offset

    def __hash__(self):
        return hash((self.id, self.type, self.offset))

    def short_repr(self):
        return f"Color({self.id})"

    def as_dict(self):
        return dataclass.as_dict(self)


@dataclass
class Abundance:
    id: str
    value: float
    type: str = "Abundance"
    encoded_color_value: int | None = None
    offset: int | None = None

    def set_offset(self, offset: int):
        self.offset = offset

    def __hash__(self):
        return hash((self.id, self.value, self.offset))

    def short_repr(self):
        return f"Abundance({self.id}, {self.value})"


@dataclass
class SubseqData:
    id: str
    type: str
    start: int
    stop: int
    list_attr: List[str]
    chr: str | None = None
    strand: str | None = None
    offset: int | None = None
    first_kmer: int | None = None
    last_kmer: int | None = None
    gene: str | None = None
    transcript: str | None = None

    def short_repr(self):
        return f"{self.type}{self.id}"

    def set_offset(self, offset: int):
        self.offset = offset

    def __hash__(self):
        return hash((self.id, self.type, self.offset))

    def add_first_kmer(self, kmer: Kmer):
        assert isinstance(kmer, Kmer)
        self.first_kmer = kmer.data
        return self

    def get_first_kmer(self, k):
        return Kmer(int.from_bytes(self.first_kmer), k)

    def get_last_kmer(self, k):
        return Kmer(int.from_bytes(self.last_kmer), k)

    def add_last_kmer(self, kmer: Kmer):
        assert isinstance(kmer, Kmer)
        self.last_kmer = kmer.data
        return self

    def __eq__(self, other):
        if type(self) is not type(other):
            return False
        if self.id != other.id:
            return False
        if self.type != other.type:
            return False
        return True

    def as_dict(self):
        return dataclass.as_dict(self)


class Zero:
    """Abstract symbol to be used to be encoded as the INT 0.
    Should be used for filtering between kmer and non kmer metadata.
    e.g. G.find_all_nodes(lambda e:e.lt(Zero())) will return all nodes
    with some metadata.

    Remark that it works as BLOB type in SQLite are always larger than 0 apparently.

    """


Metadata = Color | SubseqData
ViziKey = Metadata | Literal["sequence"] | Kmer | Zero


def metadata_from_dict(d: dict) -> Metadata:
    if d["type"] == "Color":
        return Color(**d)
    return SubseqData(**d)


def encode_kmer(kmer: Kmer, k: int) -> int | bytes:
    data = kmer.data
    if k < 32:
        return data
    if k < 64:
        return int.to_bytes(data, length=math.ceil(k / 4))
    assert isinstance(data, bytes)
    return data


def decode_kmer(data: bytes | int, k: int) -> Kmer:
    if isinstance(data, bytes):
        if len(data) <= 16:
            return Kmer(int.from_bytes(data), k)
        return Kmer(data, k)
    return Kmer(data, k)



from vizitig import info
from vizitig.env_var import (
    VIZITIG_DEFAULT_INDEX,
    VIZITIG_PYTHON_ONLY,
    VIZITIG_PROC_NUMBER,
    VIZITIG_WORK_MEM,
)
from vizitig.cli import subparsers
from vizitig.errors import (
    NoIndex,
    UnknownIndexType,
    VizIndexError,
    NoData,
)
from vizitig.index.classes import (
    Shard,
    IndexInfo,
    GraphIndex,
    TemporaryKmerSet,
    TemporaryKmerIndex,
    index_path_name,
)
from vizitig.index.sqlite_index import SQLiteIndex

if not VIZITIG_PYTHON_ONLY:
    from vizitig.index.rust_index import RustIndex
from vizitig.utils import sizeof_fmt
from vizitig.types import Kmer, DNA, AvailableKmerSize

from typing import Callable, Iterator, Tuple


__all__ = [
    "load_kmer_index",
    "index_info",
    "build_kmer_index",
    "drop_kmer_index",
    "SQLiteIndex",
    "RustIndex",
]

index_types: list[str] = sorted(
    Shard.subclasses,
    key=lambda e: Shard.subclasses[e].priority,
)


def temporary_kmerset(
    iterator: Callable[[], Iterator[Kmer]] | Callable[[], Iterator[DNA]],
    shard_number: int | None = None,
    k: int | None = None,
    index_type: str | None = None,
) -> TemporaryKmerSet:
    """Build a temporary set to store kmer."""
    if shard_number is None:
        shard_number = VIZITIG_PROC_NUMBER
    if index_type is None:
        var_env_default_index = VIZITIG_DEFAULT_INDEX
        if not var_env_default_index:
            index_type = index_types[-1]
        else:
            index_type = var_env_default_index
    IndexType = Shard.subclasses[index_type]
    try:
        first = next(iterator())
    except StopIteration:
        raise NoData

    if isinstance(first, Kmer):
        k = k or first.size
        assert k is not None
        return TemporaryKmerSet.build_kmer(IndexType, iterator, shard_number, k)
    if isinstance(first, DNA):
        assert k is not None
        print("AAAAAAAAAAAAAAAAAAAAAAAAAAAAa")
        print(shard_number)
        return TemporaryKmerSet.build_dna(IndexType, iterator, shard_number, k)

    raise ValueError(first)


def temporary_kmerindex(
    iterator: Callable[[], Iterator[Tuple[Kmer, int]]]
    | Callable[[], Iterator[Tuple[DNA, int]]],
    k: int | None = None,
    shard_number: int | None = None,
    index_type: str | None = None,
) -> TemporaryKmerIndex:
    """Build a temporary set to store a Kmer index"""
    if shard_number is None:
        shard_number = VIZITIG_PROC_NUMBER
    if index_type is None:
        var_env_default_index = VIZITIG_DEFAULT_INDEX
        if not var_env_default_index:
            index_type = index_types[-1]
        else:
            index_type = var_env_default_index
    first, _ = next(iterator())
    IndexType = Shard.subclasses[index_type]
    if isinstance(first, Kmer):
        k = k or first.size
        return TemporaryKmerIndex.build_kmer(IndexType, iterator, shard_number, k)
    if isinstance(first, DNA):
        assert k is not None
        return TemporaryKmerIndex.build_dna(IndexType, iterator, shard_number, k)

    raise ValueError(first)


def build_kmer_index(
    gname: str,
    index_type: str | None = None,
    shard_number: int | None = None,
    small_k: int | None = None,
):
    if shard_number is None:
        G = info.get_graph(gname)
        shard_number = int(
            max(
                G.metadata.k
                * G.metadata.kmer_size
                * VIZITIG_PROC_NUMBER
                / VIZITIG_WORK_MEM,
                VIZITIG_PROC_NUMBER,
            )
        )
    try:
        if index_type is None:
            var_env_default_index = VIZITIG_DEFAULT_INDEX
            if not var_env_default_index:
                index_type = index_types[-1]
            else:
                index_type = var_env_default_index
        IndexType = Shard.subclasses[index_type]
        GraphIndex.build_dna(gname, IndexType, shard_number, small_k=small_k)

    except KeyError:
        raise UnknownIndexType(index_type)


def load_kmer_index(
    gname: str, index_type: str | None = None, small_k: int | None = None
) -> GraphIndex:
    if index_type is None:
        p = len(index_types) - 1
        IndexType = Shard.subclasses[index_types[p]]
        while p >= 0:
            try:
                return GraphIndex.from_graph(gname, IndexType, small_k=small_k)
            except VizIndexError:
                p -= 1
                IndexType = Shard.subclasses[index_types[p]]
                continue
        raise NoIndex
    else:
        try:
            IndexType = Shard.subclasses[index_type]
        except KeyError:
            raise UnknownIndexType(index_type)
    return GraphIndex.from_graph(gname, IndexType, small_k)


def index_info(gname: str) -> list[IndexInfo]:
    result = []
    for index_type in index_types:
        try:
            idx = load_kmer_index(gname, index_type)
            result.append(idx.info())
        except VizIndexError:
            pass
    result.extend(small_k_index_info(gname))
    return result


def small_k_index_info(gname: str) -> list[IndexInfo]:
    result = []
    index_path = index_path_name(gname) / "small_k"
    for fsmall_k in index_path.glob("*"):
        small_k = int(str(fsmall_k.name))
        for idx_type in (index_path / fsmall_k).glob("*"):
            try:
                idx = load_kmer_index(gname, str(idx_type.name), small_k)
                result.append(idx.info())
            except VizIndexError:
                pass
    return result


def drop_kmer_index(gname: str, version: str, small_k: int | None = None):
    index = load_kmer_index(gname, version, small_k)
    index.drop()


def main(args):
    if args.action == "list":
        L = index_info(args.name)
        for index in L:
            print(f"{index.type}\t{sizeof_fmt(index.size)}")
    if args.action == "build":
        build_kmer_index(args.name, args.type, args.shard_number, small_k=args.small_k)
    if args.action == "drop":
        drop_kmer_index(args.name, args.type, small_k=args.small_k)


parser = subparsers.add_parser(
    "index",
    help="Index utilities of graph.",
    description=f"""
Index accelerate a lot many ingestion and search operation.
Default index can be setup through environnement variable VIZITIG_DEFAULT_INDEX. 
Possible values are {index_types}""",
    epilog=f"Available values for k {AvailableKmerSize}",
)

parser.set_defaults(func=main)


parser.add_argument(
    "action",
    choices=("build", "drop", "list"),
    help="The action to perform on the graph",
)

parser.add_argument(
    "name",
    help=f"Name of the graph to index: {info.graphs_list()}",
    metavar="name",
    choices=info.graphs_list(),
    type=str,
)
parser.add_argument(
    "-t",
    "--type",
    help=f"The index type: {index_types}",
    metavar="type",
    choices=index_types,
    default=None,
)

parser.add_argument(
    "--small-k",
    help="Force index on a smaller kmer value that the one provided by the graph",
    metavar="small_k",
    type=int,
    default=None,
)
parser.add_argument(
    "-s",
    "--shard_number",
    help="The number of shard (default will depend of the graph size and available RAM)",
    type=int,
    metavar="shard_nb",
    default=None,
)



from pytest import mark
from vizibridge import DNA, Kmer

from vizitig.types import DNAPython, KmerPython, Nucleotide, AvailableKmerSize

sequences_and_revcomp = [
    ("ATCGTAGCTGATCG", "CGATCAGCTACGAT"),
    ("ATCGTAGCTGATC", "GATCAGCTACGAT"),
    ("ATCGTAGCTGAT", "ATCAGCTACGAT"),
    ("ATCGTAGCTG", "CAGCTACGAT"),
    ("AAAAAAAAAA", "TTTTTTTTTT"),
    ("ACTGCATATGCAGT", "ACTGCATATGCAGT"),
    (
        "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT",
        "ATCAGCTACGATCGATCTAGCCTAGCTACTAGACTAAGCTAAGCTAATCGATCAGACTAGCAT",
    ),
]

kmers_and_canonical = [
    ("CGATCAGCTACGAT", "ATCGTAGCTGATCG"),
    ("ATCGTAGCTGATC", "ATCGTAGCTGATC"),
    ("ATCGTAGCTGAT", "ATCAGCTACGAT"),
    ("CAGCTACGAT", "ATCGTAGCTG"),
    ("AAAAAAAAAA", "AAAAAAAAAA"),
    ("ACTGCATATGCAGT", "ACTGCATATGCAGT"),
    (
        "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT",
        "ATCAGCTACGATCGATCTAGCCTAGCTACTAGACTAAGCTAAGCTAATCGATCAGACTAGCAT",
    ),
]

sequences = [
    "CGATCAGCTACGAT",
    "ATCGTAGCTGATC",
    "ATCGTAGCTGAT",
    "CAGCTACGAT",
    "AAAAAAAAAA",
    "ACTGCATATGCAGT",
    "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT",
    "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT" * 5,
    "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT" * 30,
]

kmers_add_left = [
    ("ATCG", "T", "TATC"),
    ("AAAA", "C", "CAAA"),
    ("TTTT", "G", "GTTT"),
    ("GCGC", "G", "GGCG"),
    ("CAGCTACGAT", "A", "ACAGCTACGA"),
    ("ATCGTAGCTGATC", "T", "TATCGTAGCTGAT"),
    (
        "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT",
        "T",
        "TATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGA",
    ),
]

kmers_add_right = [
    ("ATCG", "T", "TCGT"),
    ("AAAA", "C", "AAAC"),
    ("TTTT", "G", "TTTG"),
    ("GCGC", "G", "CGCG"),
    ("CAGCTACGAT", "A", "AGCTACGATA"),
    ("ATCGTAGCTGATC", "T", "TCGTAGCTGATCT"),
    (
        "ATGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGAT",
        "A",
        "TGCTAGTCTGATCGATTAGCTTAGCTTAGTCTAGTAGCTAGGCTAGATCGATCGTAGCTGATA",
    ),
]

kmer_eq = [
    ("AAAA", KmerPython(0, 4), True),
    ("AAAC", KmerPython(1, 4), True),
    ("AAAG", KmerPython(2, 4), True),
    ("AAAT", KmerPython(3, 4), True),
    ("AACA", KmerPython(4, 4), True),
    ("AACA", KmerPython(5, 4), False),
    ("TACA", KmerPython(0, 4), False),
]

kmer_types = [(KmerPython, DNAPython), (Kmer, DNA)]


class TestKmer:
    @mark.parametrize("kmer, okmer, is_eq", kmer_eq)
    def test_kmer_eq(self, kmer: str, okmer: KmerPython, is_eq: bool):
        assert (KmerPython.from_sequence(DNAPython(kmer)) == okmer) == is_eq

    @mark.parametrize("kmer, nucleotide, rkmer", kmers_add_left)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_kmer_class_add_left(
        self,
        kmer: str,
        nucleotide: Nucleotide,
        rkmer: str,
        Kmer,
        DNA,
    ):
        K = Kmer.from_sequence(DNA(kmer))
        K = K.add_left_nucleotid(nucleotide)
        RK = Kmer.from_sequence(DNA(rkmer))
        assert K == RK

    @mark.parametrize("kmer, nucleotide, rkmer", kmers_add_right)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_kmer_class_add_right(
        self,
        kmer: str,
        nucleotide: Nucleotide,
        rkmer: str,
        Kmer,
        DNA,
    ):
        K = Kmer.from_sequence(DNA(kmer))
        K = K.add_right_nucleotid(nucleotide)
        RK = Kmer.from_sequence(DNA(rkmer))
        assert K == RK

    @mark.parametrize("seq, rseq", sequences_and_revcomp)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_kmer_class_reverse_complement(self, seq: str, rseq: str, Kmer, DNA):
        assert Kmer.from_sequence(
            DNA(seq),
        ).reverse_complement() == Kmer.from_sequence(DNA(rseq))

    @mark.parametrize("kmer, c_kmer", kmers_and_canonical)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_kmer_class_canonical(self, kmer: str, c_kmer: str, Kmer, DNA):
        assert Kmer.from_sequence(DNA(kmer)).canonical() == Kmer.from_sequence(
            DNA(c_kmer),
        )

    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_kmer_class_enum_kmer(self, Kmer, DNA):
        assert list(map(str, DNA("ACTG"))) == [i for i in "ACTG"]


class TestDNA:
    @mark.parametrize("seq", sequences)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_DNAPython_builder(self, seq: str, Kmer, DNA):
        assert str(DNA(seq)) == seq

    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_dna_class_enum_canonical_kmer(self, Kmer, DNA):
        assert list(DNA("ACTG").enum_canonical_kmer(2)) == [
            Kmer.from_sequence(DNA("AC")),
            Kmer.from_sequence(DNA("AG")),
            Kmer.from_sequence(DNA("CA")),
        ]

    @mark.parametrize("seq", sequences)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_dna_class_enum_kmer_returns_right_kmer_count(self, seq: str, Kmer, DNA):
        for m in AvailableKmerSize:
            assert len(list(DNA(seq).enum_kmer(m))) == max(0, len(seq) - m + 1)

    @mark.parametrize("seq", sequences)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_dna_class_enum_canonical_kmer_returns_right_kmer_count(
        self,
        seq: str,
        Kmer,
        DNA,
    ):
        for m in sorted(AvailableKmerSize):
            assert len(list(DNA(seq).enum_canonical_kmer(m))) == max(
                0, len(seq) - m + 1
            )

    @mark.parametrize("seq", sequences)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_dna_class_enum_valid_kmer(
        self,
        seq: str,
        Kmer,
        DNA,
    ):
        for m in sorted(AvailableKmerSize):
            for i, kmer in enumerate(DNA(seq).enum_kmer(m)):
                assert str(kmer) in seq

    @mark.parametrize("seq", sequences)
    @mark.parametrize("Kmer, DNA", kmer_types)
    def test_dna_class_canonicality_of_enum_canonical_kmer(self, seq: str, Kmer, DNA):
        for m in sorted(AvailableKmerSize):
            assert all(kmer.is_canonical() for kmer in DNA(seq).enum_canonical_kmer(m))
